// lscp_client.c

#include "lscp_client.h"
#include "lscp_parser.h"


//-------------------------------------------------------------------------
// Helper functions.

// Trimming left spaces...
static const char *_lscp_ltrim ( const char *psz )
{
    while (isspace(*psz))
        psz++;
    return psz;
}

// Result buffer internal settler.
static void _lscp_client_set_result ( lscp_client_t *pClient, const char *pszResult, int iErrno )
{
    if (pClient->pszResult)
        free(pClient->pszResult);
    pClient->pszResult = NULL;

    pClient->iErrno = iErrno;

    if (pszResult)
        pClient->pszResult = strdup(_lscp_ltrim(pszResult));
}


// Engine info struct cache member.
static void _lscp_engine_info_init ( lscp_engine_info_t *pEngineInfo )
{
    pEngineInfo->description = NULL;
    pEngineInfo->version     = NULL;
}

static void _lscp_engine_info_reset ( lscp_engine_info_t *pEngineInfo )
{
    if (pEngineInfo->description)
        free(pEngineInfo->description);
    if (pEngineInfo->version)
        free(pEngineInfo->version);

    _lscp_engine_info_init(pEngineInfo);
}


// Channel info struct cache member.
static void _lscp_channel_info_init ( lscp_channel_info_t *pChannelInfo )
{
    pChannelInfo->engine_name   = NULL;
    pChannelInfo->audio_type    = LSCP_AUDIO_NONE;
    pChannelInfo->audio_channel = LSCP_CHANNEL_INVALID;
    pChannelInfo->instrument    = NULL;
    pChannelInfo->midi_type     = LSCP_MIDI_NONE;
    pChannelInfo->midi_port     = NULL;
    pChannelInfo->midi_channel  = LSCP_CHANNEL_INVALID;
    pChannelInfo->volume        = 0.0;
}

static void _lscp_channel_info_reset ( lscp_channel_info_t *pChannelInfo )
{
    if (pChannelInfo->engine_name)
        free(pChannelInfo->engine_name);
    if (pChannelInfo->instrument)
        free(pChannelInfo->instrument);
    if (pChannelInfo->midi_port)
        free(pChannelInfo->midi_port);

    _lscp_channel_info_init(pChannelInfo);
}


//-------------------------------------------------------------------------
// UDP service (datagram oriented).

static void _lscp_client_udp_proc ( void *pvClient )
{
    lscp_client_t *pClient = (lscp_client_t *) pvClient;
    struct sockaddr_in addr;
    int  cAddr;
    char achBuffer[LSCP_BUFSIZ];
    int  cchBuffer;
    lscp_parser_t tok;

#ifdef DEBUG
    fprintf(stderr, "_lscp_client_udp_proc: Client waiting for events.\n");
#endif

    while (pClient->udp.iState) {
        cAddr = sizeof(struct sockaddr_in);
        cchBuffer = recvfrom(pClient->udp.sock, achBuffer, sizeof(achBuffer), 0, (struct sockaddr *) &addr, &cAddr);
        if (cchBuffer > 0) {
#ifdef DEBUG
            lscp_socket_trace("_lscp_client_udp_proc: recvfrom", &addr, achBuffer, cchBuffer);
#endif
            lscp_parser_init(&tok, achBuffer, cchBuffer);
            if (lscp_parser_test(&tok, "PING")) {
                lscp_parser_nextint(&tok);
                snprintf(achBuffer, sizeof(achBuffer) - 1, "PONG %s\r\n", lscp_parser_next(&tok));
                cchBuffer = strlen(achBuffer);
                if (sendto(pClient->udp.sock, achBuffer, cchBuffer, 0, (struct sockaddr *) &addr, cAddr) < cchBuffer)
                    lscp_socket_perror("_lscp_client_udp_proc: sendto");
#ifdef DEBUG
            fprintf(stderr, "> %s", achBuffer);
#endif
            } else {
                if ((*pClient->pfnCallback)(
                        pClient,
                        achBuffer,
                        cchBuffer,
                        pClient->pvData) != LSCP_OK) {
                    pClient->udp.iState = 0;
                }
            }
            lscp_parser_free(&tok);
        } else {
            lscp_socket_perror("_lscp_client_udp_proc: recvfrom");
            pClient->udp.iState = 0;
        }
    }

#ifdef DEBUG
    fprintf(stderr, "_lscp_client_udp_proc: Client closing.\n");
#endif
}


//-------------------------------------------------------------------------
// Client socket functions.


lscp_client_t* lscp_client_create ( char *pszHost, int iPort, lscp_client_proc_t pfnCallback, void *pvData )
{
    lscp_client_t  *pClient;
    struct hostent *pHost;
    lscp_socket_t sock;
    struct sockaddr_in addr;
    int cAddr;
    int iSockOpt = (-1);

    if (pfnCallback == NULL) {
        fprintf(stderr, "lscp_client_create: Invalid client callback function.\n");
        return NULL;
    }

    pHost = gethostbyname(pszHost);
    if (pHost == NULL) {
        lscp_socket_perror("lscp_client_create: gethostbyname");
        return NULL;
    }

    // Allocate client descriptor...

    pClient = (lscp_client_t *) malloc(sizeof(lscp_client_t));
    if (pClient == NULL) {
        fprintf(stderr, "lscp_client_create: Out of memory.\n");
        return NULL;
    }
    memset(pClient, 0, sizeof(lscp_client_t));

    pClient->pfnCallback = pfnCallback;
    pClient->pvData = pvData;

#ifdef DEBUG
    fprintf(stderr, "lscp_client_create: pClient=%p: pszHost=%s iPort=%d.\n", pClient, pszHost, iPort);
#endif

    // Prepare the TCP connection socket...

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == INVALID_SOCKET) {
        lscp_socket_perror("lscp_client_create: tcp: socket");
        free(pClient);
        return NULL;
    }

#if defined(WIN32)
    if (setsockopt(sock, SOL_SOCKET, SO_DONTLINGER, (char *) &iSockOpt, sizeof(int)) == SOCKET_ERROR)
        lscp_socket_perror("lscp_client_create: tcp: setsockopt(SO_DONTLINGER)");
#endif

#ifdef DEBUG
    lscp_socket_getopts("lscp_client_create: tcp", sock);
#endif

    cAddr = sizeof(struct sockaddr_in);
    memset((char *) &addr, 0, cAddr);
    addr.sin_family = pHost->h_addrtype;
    memmove((char *) &(addr.sin_addr), pHost->h_addr, pHost->h_length);
    addr.sin_port = htons(iPort);

    if (connect(sock, (struct sockaddr *) &addr, cAddr) == SOCKET_ERROR) {
        lscp_socket_perror("lscp_client_create: tcp: connect");
        closesocket(sock);
        free(pClient);
        return NULL;
    }

    lscp_socket_agent_init(&(pClient->tcp), sock, &addr, cAddr);

#ifdef DEBUG
    fprintf(stderr, "lscp_client_create: tcp: pClient=%p: sock=%d addr=%s port=%d.\n", pClient, pClient->tcp.sock, inet_ntoa(pClient->tcp.addr.sin_addr), ntohs(pClient->tcp.addr.sin_port));
#endif

    // Prepare the UDP datagram service socket...

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock == INVALID_SOCKET) {
        lscp_socket_perror("lscp_client_create: udp: socket");
        lscp_socket_agent_free(&(pClient->tcp));
        free(pClient);
        return NULL;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &iSockOpt, sizeof(int)) == SOCKET_ERROR)
        lscp_socket_perror("lscp_client_create: udp: setsockopt(SO_REUSEADDR)");

#ifdef DEBUG
    lscp_socket_getopts("lscp_client_create: udp", sock);
#endif

    cAddr = sizeof(struct sockaddr_in);
    memset((char *) &addr, 0, cAddr);
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(0);

    if (bind(sock, (const struct sockaddr *) &addr, cAddr) == SOCKET_ERROR) {
        lscp_socket_perror("lscp_client_create: udp: bind");
        lscp_socket_agent_free(&(pClient->tcp));
        closesocket(sock);
        free(pClient);
        return NULL;
    }

    if (getsockname(sock, (struct sockaddr *) &addr, &cAddr) == SOCKET_ERROR) {
        lscp_socket_perror("lscp_client_create: udp: getsockname");
        lscp_socket_agent_free(&(pClient->tcp));
        closesocket(sock);
        free(pClient);
        return NULL;
    }

    lscp_socket_agent_init(&(pClient->udp), sock, &addr, cAddr);

#ifdef DEBUG
    fprintf(stderr, "lscp_client_create: udp: pClient=%p: sock=%d addr=%s port=%d.\n", pClient, pClient->udp.sock, inet_ntoa(pClient->udp.addr.sin_addr), ntohs(pClient->udp.addr.sin_port));
#endif

    // Initialize cached members.
    pClient->engines = NULL;
    _lscp_engine_info_init(&(pClient->engine_info));
    _lscp_channel_info_init(&(pClient->channel_info));
    // Initialize error stuff.
    pClient->pszResult = NULL;
    pClient->iErrno = -1;
    // Stream usage stuff.
    pClient->buffer_fill = NULL;
    pClient->iStreamCount = 0;

    // Now's finally time to startup threads...
    
    // UDP service thread...
    if (lscp_socket_agent_start(&(pClient->udp), _lscp_client_udp_proc, pClient, 0) != LSCP_OK) {
        lscp_socket_agent_free(&(pClient->tcp));
        lscp_socket_agent_free(&(pClient->udp));
        free(pClient);
        return NULL;
    }

    // Finally we've some success...
    return pClient;
}


lscp_status_t lscp_client_join ( lscp_client_t *pClient )
{
    if (pClient == NULL)
        return LSCP_FAILED;

#ifdef DEBUG
    fprintf(stderr, "lscp_client_join: pClient=%p.\n", pClient);
#endif

//  lscp_socket_agent_join(&(pClient->udp));
    lscp_socket_agent_join(&(pClient->tcp));

    return LSCP_OK;
}


lscp_status_t lscp_client_destroy ( lscp_client_t *pClient )
{
    if (pClient == NULL)
        return LSCP_FAILED;

#ifdef DEBUG
    fprintf(stderr, "lscp_client_destroy: pClient=%p.\n", pClient);
#endif

    // Free up all cached members.
    _lscp_channel_info_reset(&(pClient->channel_info));
    _lscp_engine_info_reset(&(pClient->engine_info));
    if (pClient->engines)
        free(pClient->engines);
    pClient->engines = NULL;
    // Free result error stuff.
    _lscp_client_set_result(pClient, NULL, 0);
    // Frre stream usage stuff.
    if (pClient->buffer_fill)
        free(pClient->buffer_fill);
    pClient->buffer_fill = NULL;
    pClient->iStreamCount = 0;

    // Free socket agents.
    lscp_socket_agent_free(&(pClient->udp));
    lscp_socket_agent_free(&(pClient->tcp));

    free(pClient);

    return LSCP_OK;
}


lscp_status_t lscp_client_call ( lscp_client_t *pClient, const char *pchBuffer, int cchBuffer, char *pchResult, int *pcchResult )
{
    lscp_status_t ret = LSCP_FAILED;

    if (pClient == NULL)
        return ret;
    if (pchBuffer == NULL || cchBuffer < 1)
        return ret;
    if (pchResult == NULL || pcchResult == NULL)
        return ret;
    if (*pcchResult < 1)
        return ret;

    if (send(pClient->tcp.sock, pchBuffer, cchBuffer, 0) < cchBuffer) {
        lscp_socket_perror("lscp_client_call: send");
    } else {
        *pcchResult = recv(pClient->tcp.sock, pchResult, *pcchResult, 0);
        if (*pcchResult < 1)
            lscp_socket_perror("lscp_client_call: recv");
        else
            ret = LSCP_OK;
    }
    
    return ret;
}


lscp_status_t lscp_client_subscribe ( lscp_client_t *pClient )
{
    lscp_status_t ret = LSCP_FAILED;
    char szBuffer[LSCP_BUFSIZ];
    int cchBuffer;

    if (pClient == NULL)
        return ret;

    cchBuffer = sizeof(szBuffer) - 1;
    snprintf(szBuffer, cchBuffer, "SUBSCRIBE NOTIFICATION %d\r\n", ntohs(pClient->udp.addr.sin_port));

    ret = lscp_client_call(pClient, szBuffer, strlen(szBuffer), szBuffer, &cchBuffer);
#ifdef DEBUG
    if (ret == LSCP_OK) {
        szBuffer[cchBuffer] = (char) 0;
        fprintf(stderr, "lscp_client_subscribe: %s.\n", szBuffer);
    }
#endif
    return ret;
}


lscp_status_t lscp_client_unsubscribe ( lscp_client_t *pClient )
{
    lscp_status_t ret = LSCP_FAILED;
    char szBuffer[LSCP_BUFSIZ];
    int cchBuffer;

    if (pClient == NULL)
        return ret;

    cchBuffer = sizeof(szBuffer) - 1;
    snprintf(szBuffer, cchBuffer, "UNSUBSCRIBE NOTIFICATION %d\n\n", ntohs(pClient->udp.addr.sin_port));

    ret = lscp_client_call(pClient, szBuffer, strlen(szBuffer), szBuffer, &cchBuffer);
#ifdef DEBUG
    if (ret == LSCP_OK) {
        szBuffer[cchBuffer] = (char) 0;
        fprintf(stderr, "lscp_client_unsubscribe: %s.\n", szBuffer);
    }
#endif
    return ret;
}


//-------------------------------------------------------------------------
// Client common protocol functions.


lscp_status_t lscp_client_query ( lscp_client_t *pClient, const char *pszQuery )
{
    lscp_status_t ret = LSCP_FAILED;
    char  achResult[LSCP_BUFSIZ];
    int   cchResult;
    const char *pszSeps = ":";
    const char *pszResult;
    char *pszToken;
    char *pch;
    int   iErrno;

    if (pClient == NULL)
        return ret;

    pszResult = NULL;
    iErrno = -1;

    // Do the socket transaction...
    cchResult = sizeof(achResult);
    ret = lscp_client_call(pClient, pszQuery, strlen(pszQuery), achResult, &cchResult);
    if (ret == LSCP_OK) {
        // Always force the result to be null terminated (and trim trailing CRLF)!
        while (cchResult > 0 && (achResult[cchResult - 1] == '\n' || achResult[cchResult- 1] == '\r'))
            cchResult--;
        achResult[cchResult] = (char) 0;
        // Check if the response is an error or warning message.
        if (strncmp(achResult, "ERR:", 4) == 0)
            ret = LSCP_ERROR;
        else if (strncmp(achResult, "WRN:", 4) == 0)
            ret = LSCP_WARNING;
        // So we got a result...
        if (ret == LSCP_OK) {
            pszResult = achResult;
            iErrno = 0;
        } else {
            // Parse the error/warning message, skip first colon...
            pszToken = strtok_r(achResult, pszSeps, &(pch));
            if (pszToken) {
                // Get the error number...
                pszToken = strtok_r(NULL, pszSeps, &(pch));
                if (pszToken) {
                    iErrno = atoi(pszToken);
                    // And make the message text our final result.
                    pszResult = strtok_r(NULL, pszSeps, &(pch));
                }
            }
        }
    }

    // Make the result official...
    _lscp_client_set_result(pClient, pszResult, iErrno);

    return ret;
}


// Query result accessors.
const char *lscp_client_get_result ( lscp_client_t *pClient )
{
    if (pClient == NULL)
        return NULL;

    return pClient->pszResult;
}


int lscp_client_get_errno ( lscp_client_t *pClient )
{
    if (pClient == NULL)
        return -1;

    return pClient->iErrno;
}


//-------------------------------------------------------------------------
// Client command protocol functions.


lscp_status_t lscp_load_instrument ( lscp_client_t *pClient, const char *pszFileName, int iIndex, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];
    
    if (pszFileName == NULL || iChannel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "LOAD INSTRUMENT %s %d %d\r\n", pszFileName, iIndex, iChannel);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_load_engine ( lscp_client_t *pClient, const char *pszEngineName, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];

    if (pszEngineName == NULL || iChannel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "LOAD ENGINE %s %d\r\n", pszEngineName, iChannel);
    return lscp_client_query(pClient, szQuery);
}


int lscp_get_channels ( lscp_client_t *pClient )
{
    int iChannels = -1;
    if (lscp_client_query(pClient, "GET CHANNELS\r\n") == LSCP_OK)
        iChannels = atoi(lscp_client_get_result(pClient));
    return iChannels;
}


lscp_status_t lscp_add_channel ( lscp_client_t *pClient )
{
    return lscp_client_query(pClient, "ADD CHANNEL\r\n");
}


lscp_status_t lscp_remove_channel ( lscp_client_t *pClient, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "REMOVE CHANNEL %d\r\n", iChannel);
    return lscp_client_query(pClient, szQuery);
}


const char **lscp_get_available_engines ( lscp_client_t *pClient )
{
    const char *pszResult;
    const char **ppEngines = NULL;
    const char  *pszCrlf = "\r\n";
    int   iEngines = 0;
    char *pszToken;
    char *pch;

    if (pClient->engines == NULL)
        pClient->engines = (const char **) malloc(LSCP_MAX_ENGINES * sizeof(const char *));
    if (pClient->engines == NULL)
        return NULL;

    if (lscp_client_query(pClient, "GET AVAILABLE_ENGINES\r\n") == LSCP_OK) {
        pszResult = lscp_client_get_result(pClient);
        ppEngines = pClient->engines;
        pszToken = strtok_r((char *) pszResult, pszCrlf, &(pch));
        while (pszToken && iEngines < LSCP_MAX_ENGINES) {
            if (*pszToken) {
                ppEngines[iEngines] = pszToken;
                iEngines++;
            }
            pszToken = strtok_r(NULL, pszCrlf, &(pch));
        }
        ppEngines[iEngines] = NULL;
    }

    return ppEngines;
}


lscp_engine_info_t *lscp_get_engine_info ( lscp_client_t *pClient, const char *pszEngineName )
{
    lscp_engine_info_t *pEngineInfo = NULL;
    char szQuery[LSCP_BUFSIZ];
    const char *pszResult;
    const char *pszSeps = ":";
    const char *pszCrlf = "\r\n";
    char *pszToken;
    char *pch;

    if (pszEngineName == NULL)
        return NULL;

    snprintf(szQuery, sizeof(szQuery), "GET ENGINE INFO %s\r\n", pszEngineName);
    if (lscp_client_query(pClient, szQuery) == LSCP_OK) {
        pszResult = lscp_client_get_result(pClient);
        pEngineInfo = &(pClient->engine_info);
        _lscp_engine_info_reset(pEngineInfo);
        pszToken = strtok_r((char *) pszResult, pszSeps, &(pch));
        while (pszToken) {
            if (strcmp(pszToken, "DESCRIPTION") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pEngineInfo->description = strdup(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "VERSION") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pEngineInfo->version = strdup(_lscp_ltrim(pszToken));
            }
            pszToken = strtok_r(NULL, pszSeps, &(pch));
        }
    }

    return pEngineInfo;
}


lscp_channel_info_t *lscp_get_channel_info ( lscp_client_t *pClient, int iChannel )
{
    lscp_channel_info_t *pChannelInfo = NULL;
    char szQuery[LSCP_BUFSIZ];
    const char *pszResult;
    const char *pszSeps = ":";
    const char *pszCrlf = "\r\n";
    char *pszToken;
    char *pch;

    if (iChannel < 0)
        return NULL;
        
    snprintf(szQuery, sizeof(szQuery), "GET CHANNEL INFO %d\r\n", iChannel);
    if (lscp_client_query(pClient, szQuery) == LSCP_OK) {
        pszResult = lscp_client_get_result(pClient);
        pChannelInfo = &(pClient->channel_info);
        _lscp_channel_info_reset(pChannelInfo);
        pszToken = strtok_r((char *) pszResult, pszSeps, &(pch));
        while (pszToken) {
            if (strcmp(pszToken, "ENGINE_NAME") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->engine_name = strdup(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "AUDIO_OUTPUT_TYPE") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken) {
                    pszToken = (char *) _lscp_ltrim(pszToken);
                    if (strcmp(pszToken, "ALSA") == 0)
                        pChannelInfo->audio_type = LSCP_AUDIO_ALSA;
                    else if (strcmp(pszToken, "JACK") == 0)
                        pChannelInfo->audio_type = LSCP_AUDIO_JACK;
                }
            }
            else if (strcmp(pszToken, "AUDIO_OUTPUT_CHANNEL") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->audio_channel = atoi(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "INSTRUMENT") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->instrument = strdup(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "MIDI_INPUT_TYPE") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken) {
                    pszToken = (char *) _lscp_ltrim(pszToken);
                    if (strcmp(pszToken, "ALSA") == 0)
                        pChannelInfo->midi_type = LSCP_MIDI_ALSA;
                }
            }
            else if (strcmp(pszToken, "MIDI_INPUT_PORT") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->midi_port = strdup(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "MIDI_INPUT_CHANNEL") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->midi_channel = atoi(_lscp_ltrim(pszToken));
            }
            else if (strcmp(pszToken, "VOLUME") == 0) {
                pszToken = strtok_r(NULL, pszCrlf, &(pch));
                if (pszToken)
                    pChannelInfo->volume = atof(_lscp_ltrim(pszToken));
            }
            pszToken = strtok_r(NULL, pszSeps, &(pch));
        }
    }

    return pChannelInfo;
}


int lscp_get_channel_voice_count ( lscp_client_t *pClient, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];
    int iVoiceCount = -1;

    if (iChannel < 0)
        return iVoiceCount;

    snprintf(szQuery, sizeof(szQuery), "GET CHANNEL VOICE_COUNT %d\r\n", iChannel);
    if (lscp_client_query(pClient, szQuery) == LSCP_OK)
        iVoiceCount = atoi(lscp_client_get_result(pClient));

    return iVoiceCount;
}


int lscp_get_channel_stream_count ( lscp_client_t *pClient, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];
    int iStreamCount = -1;

    if (iChannel < 0)
        return iStreamCount;

    snprintf(szQuery, sizeof(szQuery), "GET CHANNEL STREAM_COUNT %d\r\n", iChannel);
    if (lscp_client_query(pClient, szQuery) == LSCP_OK)
        iStreamCount = atoi(lscp_client_get_result(pClient));

    return iStreamCount;
}


lscp_buffer_fill_t *lscp_get_channel_buffer_fill ( lscp_client_t *pClient, lscp_usage_t usage_type, int iChannel )
{
    lscp_buffer_fill_t *pBufferFill = NULL;
    char szQuery[LSCP_BUFSIZ];
    int iStreamCount = pClient->iStreamCount;
    const char *pszUsageType = (usage_type == LSCP_USAGE_BYTES ? "BYTES" : "PERCENTAGE");
    const char *pszResult;
    const char *pszSeps = "[]%,";
    char *pszToken;
    char *pch;
    int   iStream;

    if (iChannel < 0)
        return NULL;
    if (iStreamCount < 1)
        iStreamCount = lscp_get_channel_stream_count(pClient, iChannel);
    if (iStreamCount < 1)
        return NULL;

    if (pClient->iStreamCount != iStreamCount) {
        if (pClient->buffer_fill)
            free(pClient->buffer_fill);
        pClient->iStreamCount = 0;
        pClient->buffer_fill = (lscp_buffer_fill_t *) malloc(iStreamCount * sizeof(lscp_buffer_fill_t));
        pClient->iStreamCount = iStreamCount;
    }

    snprintf(szQuery, sizeof(szQuery), "GET CHANNEL BUFFER_FILL %d %d\r\n", pszUsageType, iChannel);
    if (lscp_client_query(pClient, szQuery) == LSCP_OK) {
        pszResult = lscp_client_get_result(pClient);
        pBufferFill = pClient->buffer_fill;
        pszToken = strtok_r((char *) pszResult, pszSeps, &(pch));
        iStream = 0;
        while (pszToken && iStream < pClient->iStreamCount) {
            if (*pszToken) {
                pBufferFill[iStream].stream_id = atol(pszToken);
                pszToken = strtok_r(NULL, pszSeps, &(pch));
                if (pszToken == NULL)
                    break;
                pBufferFill[iStream].stream_usage = atol(pszToken);
                iStream++;
            }
            pszToken = strtok_r(NULL, pszSeps, &(pch));
        }
    }

    return pBufferFill;
}


lscp_status_t lscp_set_channel_audio_type ( lscp_client_t *pClient, int iChannel, lscp_audio_t audio_type )
{
    char szQuery[LSCP_BUFSIZ];
    const char *pszAudioType;

    if (iChannel < 0)
        return LSCP_FAILED;

    switch (audio_type) {
    case LSCP_AUDIO_ALSA:
        pszAudioType = "ALSA";
        break;
    case LSCP_AUDIO_JACK:
        pszAudioType = "JACK";
        break;
    default:
        return LSCP_FAILED;
    }

    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL AUDIO_OUTPUT_TYPE %d %s\r\n", iChannel, pszAudioType);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_set_channel_audio_channel ( lscp_client_t *pClient, int iChannel, int audio_channel )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0 || audio_channel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL AUDIO_OUTPUT_CHANNEL %d %d\r\n", iChannel, audio_channel);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_set_channel_midi_type ( lscp_client_t *pClient, int iChannel, lscp_midi_t midi_type )
{
    char szQuery[LSCP_BUFSIZ];
    const char *pszMidiType;

    if (iChannel < 0)
        return LSCP_FAILED;

    switch (midi_type) {
    case LSCP_MIDI_ALSA:
        pszMidiType = "ALSA";
        break;
    default:
        return LSCP_FAILED;
    }
    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL MIDI_INPUT_TYPE %d %s\r\n", iChannel, pszMidiType);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_set_channel_midi_port ( lscp_client_t *pClient, int iChannel, const char *midi_port )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0 || midi_port == NULL)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL MIDI_INPUT_PORT %d %s\r\n", iChannel, midi_port);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_set_channel_midi_channel ( lscp_client_t *pClient, int iChannel, int midi_channel )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0 || midi_channel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL MIDI_INPUT_CHANNEL %d %d\r\n", iChannel, midi_channel);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_set_channel_volume ( lscp_client_t *pClient, int iChannel, float volume )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0 || volume < 0.0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "SET CHANNEL VOLUME %d %g\r\n", iChannel, volume);
    return lscp_client_query(pClient, szQuery);
}


lscp_status_t lscp_reset_channel ( lscp_client_t *pClient, int iChannel )
{
    char szQuery[LSCP_BUFSIZ];

    if (iChannel < 0)
        return LSCP_FAILED;

    snprintf(szQuery, sizeof(szQuery), "RESET CHANNEL %d\r\n", iChannel);
    return lscp_client_query(pClient, szQuery);
}


// end of lscp_socket.c
